﻿namespace David.Academy.ProgrammingInCSharp.OrcsAndDungeons
{
	/// <summary>
	/// Базов клас за единица (играч или чудовище)
	/// </summary>
	internal abstract class Unit
	{
		private readonly Game _game;
		private readonly int _damage;
		private readonly char _symbol;

		public delegate void UnitDiedHandler(Unit unit);

		/// <summary>
		/// Координати на единицата
		/// </summary>
		public Point Point { get; private set; }

		/// <summary>
		/// Оставащи точки живот на единицата
		/// </summary>
		public int HitPoints { get; private set; }

		/// <summary>
		/// Поражения, които единицата нанася при атака
		/// </summary>
		public int Damage
		{
			get { return _damage; }
		}

		/// <summary>
		/// Символ за визуализация на единицата на екрана
		/// </summary>
		public char Symbol
		{
			get { return _symbol; }
		}

		/// <summary>
		/// Указва дали единицата е мъртва
		/// </summary>
		public bool IsDead
		{
			get { return (HitPoints <= 0); }
		}

		/// <summary>
		/// Предизвиква се непосредствено след смъртта на единицата
		/// </summary>
		public event UnitDiedHandler UnitDied;

		/// <summary>
		/// Създава екземпляр на класа за единица
		/// </summary>
		/// <param name="game">Игра, част от която е единицата</param>
		/// <param name="hitPoints">Начален брой точки живот</param>
		/// <param name="damage">Щети, които нанася при атака единицата</param>
		/// <param name="symbol">Символ за визуализация на единицата на екрана</param>
		protected Unit(Game game, int hitPoints, int damage, char symbol)
		{
			_game = game;
			HitPoints = hitPoints;
			_damage = damage;
			_symbol = symbol;

			InitializeCoords();
		}

		/// <summary>
		/// Прави опит за придвижване на играча в указана посока
		/// </summary>
		/// <param name="direction">Посока, в която да се придвижи единицата</param>
		/// <returns>true, ако придвижването е било успешно</returns>
		public bool MoveOrAttack(Direction direction)
		{
			if (IsDead)
				return false;

			Point newPoint = Point.Move(direction);
			if (!_game.IsTileEmpty(newPoint))
				return false;

			Unit unit = _game.GetTileUnit(newPoint);
			if (unit != null)
			{
				if (unit.IsDead)
					return false;

				if ((this is Player && unit is Monster)
					|| (this is Monster && unit is Player))
				{
					unit.SufferDamage(Damage);
					return true;
				}
			}

			Point = newPoint;
			return true;
		}

		/// <summary>
		/// Нанася щети на единицата
		/// </summary>
		/// <param name="damage">Точки щети, които да бъдат нанесени</param>
		public void SufferDamage(int damage)
		{
			if (IsDead)
				return;

			HitPoints -= damage;
			if (HitPoints < 0)
				HitPoints = 0;

			if (IsDead && UnitDied != null)
				UnitDied(this);
		}

		/// <summary>
		/// Игра, част от която е единицата
		/// </summary>
		protected Game Game
		{
			get { return _game; }
		}

		/// <summary>
		/// Инициализира координатите на единицата с валидни стойности,
		/// така че да не се намира върху стена или друга единица
		/// </summary>
		private void InitializeCoords()
		{
			do
			{
				int x = Game.Random.Next(1, Game.Width - 1);
				int y = Game.Random.Next(1, Game.Height - 1);
				Point = new Point(x, y);
			}
			while (!Game.IsTileEmpty(Point));
		}
	}
}
